@@ -27,10 +27,6 @@
 #include "create_tmp_table.h"
 #include "sql_parse.h"
 
-#ifndef DBUG_OFF
-int dbug_json_check_min_stack_requirement();
-#endif
-
 #define HA_ERR_JSON_TABLE (HA_ERR_LAST+1)
 
 class table_function_handlerton
@@ -42,6 +38,7 @@ public:
     bzero(&m_hton, sizeof(m_hton));
     m_hton.tablefile_extensions= hton_no_exts;
     m_hton.slot= HA_SLOT_UNDEF;
+    m_hton.flags= HTON_HIDDEN;
   }
 };
 
@@ -107,9 +104,13 @@ int get_disallowed_table_deps_for_list(MEM_ROOT *mem_root,
   List_iterator<TABLE_LIST> li(*join_list);
 
   DBUG_EXECUTE_IF("json_check_min_stack_requirement",
-                  return -dbug_json_check_min_stack_requirement(););
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
   if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
-    return -1;
+    return 1;
 
   while ((table= li++))
   {
@@ -229,6 +230,10 @@ public:
   int open(const char *name, int mode, uint test_if_locked) override
   { return 0; }
   int close(void) override { return 0; }
+  void update_optimizer_costs(OPTIMIZER_COSTS *costs) override
+  {
+    memcpy(costs, &heap_optimizer_costs, sizeof(*costs));
+  }
   int rnd_init(bool scan) override;
   int rnd_next(uchar *buf) override;
   int rnd_pos(uchar * buf, uchar *pos) override;
@@ -365,8 +370,6 @@ int ha_json_table::rnd_init(bool scan)
 
 static void store_json_in_field(Field *f, const json_engine_t *je)
 {
-  String res_tmp("", 0, je->s.cs);
-
   switch (je->value_type)
   {
   case JSON_VALUE_NULL:
@@ -387,8 +390,7 @@ static void store_json_in_field(Field *f, const json_engine_t *je)
   default:
     break;
   };
-  st_append_json(&res_tmp, je->s.cs, je->value, je->value_len);
-  f->store((const char *) res_tmp.ptr(), (uint32) res_tmp.length(), je->s.cs);
+  f->store((const char *) je->value, (uint32) je->value_len, je->s.cs);
 }
 
 
@@ -711,8 +713,6 @@ TABLE *Create_json_table::start(THD *thd,
   if (!(table= Create_tmp_table::start(thd, param, table_alias)))
     DBUG_RETURN(0);
   share= table->s;
-  share->db= any_db;
-  share->table_name= { STRING_WITH_LEN("json_table") };
   share->not_usable_by_query_cache= FALSE;
   share->db_plugin= NULL;
   if (!(table->file= new (&table->mem_root) ha_json_table(share, jt)))
@@ -970,9 +970,8 @@ int Json_table_column::print(THD *thd, Field **f, String *str)
 
     (*f)->sql_type(column_type);
 
-    if ((m_format_json ? str->append(STRING_WITH_LEN(" JSON ")) : str->append(column_type)))
-      return 1;
-    if (((*f)->has_charset() && m_explicit_cs &&
+    if (str->append(column_type) ||
+        ((*f)->has_charset() && m_explicit_cs &&
          (str->append(STRING_WITH_LEN(" CHARSET ")) ||
           str->append(&m_explicit_cs->cs_name) ||
           (Charset(m_explicit_cs).can_have_collate_clause() &&
@@ -1036,7 +1035,8 @@ int Json_table_column::On_response::respond(Json_table_column *jc, Field *f,
       return 1;
     case Json_table_column::RESPONSE_DEFAULT:
       f->set_notnull();
-      m_default->save_in_field(f, TRUE);
+      f->store(m_default.str,
+          m_default.length, jc->m_defaults_cs);
       break;
   }
   return 0;
@@ -1046,11 +1046,7 @@ int Json_table_column::On_response::respond(Json_table_column *jc, Field *f,
 int Json_table_column::On_response::print(const char *name, String *str) const
 {
   LEX_CSTRING resp;
-
-  char valbuf[512];
-  String val(valbuf, sizeof(valbuf), str->charset());
-  String *ds= NULL;
-
+  const LEX_CSTRING *ds= NULL;
   if (m_response == Json_table_column::RESPONSE_NOT_SPECIFIED)
     return 0;
 
@@ -1065,7 +1061,7 @@ int Json_table_column::On_response::print(const char *name, String *str) const
     case Json_table_column::RESPONSE_DEFAULT:
     {
       lex_string_set3(&resp, STRING_WITH_LEN("DEFAULT"));
-      ds= m_default->val_str(&val);
+      ds= &m_default;
       break;
     }
     default:
@@ -1074,14 +1070,9 @@ int Json_table_column::On_response::print(const char *name, String *str) const
   }
 
   return (str->append(' ') || str->append(resp)  ||
-          (ds &&
-            (str->append(' ') ||
-             (m_default->result_type()==STRING_RESULT && str->append('\''))||
-
-             str->append_for_single_quote(ds) ||
-
-             (m_default->result_type()==STRING_RESULT && str->append('\''))))||
-
+          (ds && (str->append(STRING_WITH_LEN(" '")) ||
+                  str->append_for_single_quote(ds->str, ds->length) ||
+                  str->append('\''))) ||
           str->append(STRING_WITH_LEN(" ON ")) ||
           str->append(name, strlen(name)));
 }
@@ -1338,20 +1329,21 @@ void Table_function_json_table::fix_after_pullout(TABLE_LIST *sql_table,
 /*
   @brief
      Recursively make all tables in the join_list also depend on deps.
-
-  @return - boolean - true if error (out of memory).
 */
 
-static bool add_extra_deps(List<TABLE_LIST> *join_list, table_map deps)
+static void add_extra_deps(List<TABLE_LIST> *join_list, table_map deps)
 {
   TABLE_LIST *table;
   List_iterator<TABLE_LIST> li(*join_list);
 
   DBUG_EXECUTE_IF("json_check_min_stack_requirement",
-                  dbug_json_check_min_stack_requirement(); return true;);
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
   if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
-    return true;
-
+    return;
   while ((table= li++))
   {
     table->dep_tables |= deps;
@@ -1359,11 +1351,9 @@ static bool add_extra_deps(List<TABLE_LIST> *join_list, table_map deps)
     if ((nested_join= table->nested_join))
     {
        // set the deps inside, too
-       if (add_extra_deps(&nested_join->join_list, deps))
-         return true;
+       add_extra_deps(&nested_join->join_list, deps);
     }
   }
-  return false;
 }
 
 
@@ -1431,29 +1421,25 @@ static bool add_extra_deps(List<TABLE_LIST> *join_list, table_map deps)
   @param  join_list    List of tables to process. Initial invocation should
                        supply the JOIN's top-level table list.
   @param  nest_tables  Bitmap of all tables in the join list.
-  @param  error        Pointer to value which is set to true on stack overrun
-                       error.
 
-  @return Bitmap of all outside references that tables in join_list have,
-    or 0 on out of stack overrun error (in addition to *error= true).
+  @return Bitmap of all outside references that tables in join_list have
 */
 
 table_map add_table_function_dependencies(List<TABLE_LIST> *join_list,
-                                          table_map nest_tables,
-					  bool *error)
+                                          table_map nest_tables)
 {
   TABLE_LIST *table;
   table_map res= 0;
   List_iterator<TABLE_LIST> li(*join_list);
 
   DBUG_EXECUTE_IF("json_check_min_stack_requirement",
-                  if (dbug_json_check_min_stack_requirement())
-		    { *error= true; return 0; });
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
   if ((res=check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL)))
-  {
-    *error= true;
-    return 0;
-  }
+    return res;
 
   // Recursively compute extra dependencies
   while ((table= li++))
@@ -1462,9 +1448,7 @@ table_map add_table_function_dependencies(List<TABLE_LIST> *join_list,
     if ((nested_join= table->nested_join))
     {
       res |= add_table_function_dependencies(&nested_join->join_list,
-                                             nested_join->used_tables, error);
-      if (*error)
-	return 0;
+                                             nested_join->used_tables);
     }
     else if (table->table_function)
     {
@@ -1475,13 +1459,7 @@ table_map add_table_function_dependencies(List<TABLE_LIST> *join_list,
   res= res & ~nest_tables & ~PSEUDO_TABLE_BITS;
   // Then, make all "peers" have them:
   if (res)
-  {
-    if (add_extra_deps(join_list,  res))
-    {
-      *error= true;
-      return 0;
-    }
-  }
+    add_extra_deps(join_list,  res);
 
   return res;
 }
